/**
 * @file
 * This header is provided for compatibility only and will be removed
 * on next major bump
 */

#ifndef AVCODEC_OPT_H
#define AVCODEC_OPT_H

#include "libavcodec/version.h"

#if FF_API_OPT_H
#include "libavutil/opt.h"
#endif

#endif /* AVCODEC_OPT_H */
